using System;
using System.IO;
using System.Collections.Generic;
using Kaitai;

namespace ResourceExtractor {
    class Program {
        private static string[] FILE_EXTENSIONS = new string[] { "exe", "dll", "ocx", "sys", "scr", "mui" };

        private const string USAGE = @"Usage: {0} path [path...]

Extract HTML-type resources from every Win32 PE file found in one or more directories.
Creates a hierarchy of sub-directories under the current working directory: each file with HTML resources gets a directory, with one sub-directory per language code, holding all the found files.
For example, a single HTML file could be found at %CD%\Program Files\Something\file.dll\hello.htm\1033.";

        private static void PrintUsage() {
            Console.WriteLine(
                USAGE,
                Environment.GetCommandLineArgs()[0]
            );
        }

        static bool ExtractAll(string root, string extension, string destination) {
            bool success = false;
            foreach (string file in Directory.GetFiles(root, String.Format("*.{0}", extension), SearchOption.AllDirectories)) {
                try {
                    success |= ExtractFile(file, Path.Combine(destination, file.Substring(Path.GetPathRoot(file).Length)));
                } catch (Exception e) {
                    Console.Error.WriteLine("Could not read file {0}: {1}", file, e.Message);
                }
            }
            return success;
        }

        static bool ExtractFile(string filePath, string destination) {
            bool success = false;
            MicrosoftPe pe = MicrosoftPe.FromFile(filePath);
            foreach (MicrosoftPe.Section section in pe.Pe.Sections) {
                if (section.ResourceData == null)
                    continue;
                foreach (MicrosoftPe.ResourceDirectoryEntry entry in section.ResourceData.Entries) {
                    if (entry.Id == MicrosoftPe.ResourceDirectoryEntry.ResourceType.Html) {
                        success |= ExtractDirectoryEntry(entry, destination);
                    }
                }
            }
            return success;
        }

        static bool ExtractDirectoryEntry(MicrosoftPe.ResourceDirectoryEntry entry, string destination) {
            if (!entry.DataIsDirectory) {
                return ExtractDataEntry(entry.Data, destination);
            }
            bool success = false;
            foreach (MicrosoftPe.ResourceDirectoryEntry subEntry in entry.Directory.Entries) {
                try {
                    string subDir = subEntry.NameIsString ? subEntry.Name.Text : (subEntry.NameOffset & 65535).ToString();
                    success |= ExtractDirectoryEntry(subEntry, Path.Combine(destination, subDir));
                } catch (Exception e) {
                    Console.Error.WriteLine("Could not extract a resource directory entry to {0}: {1}", destination, e.Message);
                }
            }
            return success;
        }

        static bool ExtractDataEntry(MicrosoftPe.ResourceDataEntry entry, string destination) {
            try {
                string dirName = Path.GetDirectoryName(destination);
                if (!Directory.Exists(dirName))
                    Directory.CreateDirectory(dirName);
                File.WriteAllBytes(destination, entry.Content);
            } catch (Exception e) {
                Console.Error.WriteLine("Could not extract a resource data entry to {0}: {1}", destination, e.Message);
                return false;
            }
            return true;
        }

        static int Main(string[] args) {
            if (args.Length == 0) {
                PrintUsage();
                return 2;
            }
            List<string> parsedPaths = new List<string>();
            foreach (string arg in args) {
                try {
                    parsedPaths.Add(Path.GetFullPath(arg));
                } catch (Exception e) {
                    Console.Error.WriteLine("Could not parse path {0}: {1}", arg, e.Message);
                    return 1;
                }
            }

            // Not finding any files, or failing to read or extract any file, should return in a non-zero return code.
            bool success = false;
            foreach (string path in parsedPaths) {
                foreach (string extension in FILE_EXTENSIONS) {
                    try {
                        success |= ExtractAll(path, extension, Environment.CurrentDirectory);
                    } catch (Exception e) {
                        Console.Error.WriteLine("Could not find {0} files in {1}: {2}", extension.ToUpper(), path, e.Message);
                    }
                }
            }
            return success ? 1 : 0;
        }
    }
}
